var allNoteOptions = ["pin", "list", "sort", "dismissed"];
var allNoteOptionsLabels = ["<small>&#x1F4CC;</small> Pin", "<small>&#x1F4CB;</small> Emphasize and remove duplicate list items", "<small>&#x1F524;</small> Sort list items alphabetically", "<small>&#x2611;</small> Show dismissed list items"];

var socket = null;
var PauseTimeout = null;
var ReconnectDelay = -1;

var AuthKey = "";
var Pinned = {};
var Notebooks = {};
var NotebooksModified = 0;

var ViewNotebook = "";
var ViewNote = "";

var NoteMode = "view";
var NewNoteMode = "";
var NewViewNotebook = "";
var NewViewNote = "";
var NoteTitle = "";
var NoteBody = "";
var NoteBodyDraft = "";
var NoteCheckbox = 0;
var NoteCheckingItem = false;
var NoteHideDismissed = true;

var reconnectTimeout;
var connected = false;

var db;
var regexpListItem = /^([\s]+)?[\-*#](?![^\S\r\n]+[\-*#])(([^\S\r\n]+\[[ xX]])|[^\S\r\n])/g;
var blankOptionsDirective = "[//]: # ()\n\n";
var markdownRenderer = new marked.Renderer();
markdownRenderer.checkbox = function (chkd) {
    if (Object.keys(Notebooks).length === 0 || ViewNotebook === "" || ViewNote === "") {
        return '<input type="checkbox">';
    }

    var checked = chkd ? ' checked' : '';
    var ret = '<input type="checkbox" id="chk_' + NoteCheckbox + '" onclick="checkItem(' + NoteCheckbox + ', ' + Notebooks[ViewNotebook].Notes[ViewNote].ModifiedAt + ');return false;"' + checked + '> ';

    NoteCheckbox++;
    return ret;
};

function RunStick() {
    db = new Dexie("stick");
    db.version(1).stores({credentials: "authkey", notebooks: "id,label", notes: "notebook,name,modified,body"});

    db.credentials.each(credential => {
        AuthKey = credential.authkey;
    }).then(function () {
        Initialize();
    });
}

function Initialize() {
    var processedHash = false;
    if (window.location.hash.substring(0, 7) === '#login/') {
        processHash();
        processedHash = true;
    }

    $(window).on("focus blur", function (e) {
        var prevType = $(this).data("prevType");

        if (prevType !== e.type) {   //  reduce double fire issues
            switch (e.type) {
                case "blur":
                    if (PauseTimeout === null) {
                        PauseTimeout = setTimeout(PauseStick, 5000);
                    }
                    break;
                case "focus":
                    ResumeStick();
                    break;
            }
        }

        $(this).data("prevType", e.type);
    });

    var navigateTo = window.location.hash;
    history.replaceState({}, "stick", "#");
    history.pushState({}, "stick", "#");

    ResumeStick();

    if (!processedHash) {
        if (navigateTo !== '' && navigateTo !== '#') {
            history.pushState({}, "stick", navigateTo);
        }

        processHash();
    }

    $(window).on("popstate", function (e) {
        processHash();
    });
}

function PauseStick() {
    ReconnectDelay = -1;

    clearTimeout(PauseTimeout);
    PauseTimeout = null;

    NoteCheckingItem = false;
    if (socket != null) {
        socket.close();
        socket = null;
    }
}

function ResumeStick() {
    if (PauseTimeout !== null) {
        clearTimeout(PauseTimeout);
        PauseTimeout = null;
    }
    if (AuthKey == "") {
        AuthKey = "public";
    }
    if (webSocketReady() || ReconnectDelay !== -1) {
        return;
    }

    ReconnectDelay = 0;
    ConnectStick();

}

function ConnectStick() {
    reconnectTimeout = null;
    if (webSocketReady() || ReconnectDelay === -1) {
        return;
    }

    var loc = window.location, wsurl, pathname;
    if (loc.protocol === "https:") {
        wsurl = "wss:";
    } else {
        wsurl = "ws:";
    }
    if (loc.pathname && loc.pathname !== "") {
        pathname = loc.pathname;
    } else {
        pathname = "/";
    }
    wsurl += "//" + loc.host + pathname + "w/" + AuthKey + "/" + NotebooksModified;

    socket = new WebSocket(wsurl);
    socket.onerror = function (e) {
        console.log(e);
    };
    socket.onopen = function (e) {
        if (reconnectTimeout != null) {
            clearTimeout(reconnectTimeout);
        }
    };
    socket.onmessage = function (e) {
        if (ReconnectDelay > 0) {
            ReconnectDelay = 0;
        }

        var incomingmsg = e.data;
        var nowconnected = false;
        try {
            var data = jQuery.parseJSON(incomingmsg);
            if (data.hasOwnProperty("status") && data.status === "fail") {
                var message = 'Error';
                if (data.hasOwnProperty("message")) {
                    message += ': ' + data.message;
                }
                alert(message);
            }
            if (data.hasOwnProperty("pinned")) {
                if (data.hasOwnProperty("delta")) {
                    for (var nbid in data.pinned) {
                        for (var i in data.pinned[nbid]) {
                            var nid = data.pinned[nbid][i];

                            if (Pinned[nbid] === undefined) {
                                Pinned[nbid] = [nid];
                            } else if (!Pinned[nbid].includes(nid)) {
                                Pinned[nbid].push(nid);
                            }
                        }
                    }
                } else {
                    Pinned = data.pinned;
                }
            }
            if (data.hasOwnProperty("notebooks")) {
                var mustRender = false;
                nowconnected = true;

                if (data.hasOwnProperty("delta")) {
                    for (var nbid in data.notebooks) {
                        for (var nid in data.notebooks[nbid].Notes) {
                            Notebooks[nbid].Notes[nid] = data.notebooks[nbid].Notes[nid];

                            if (Notebooks[nbid].Notes[nid].ModifiedAt > NotebooksModified) {
                                NotebooksModified = Notebooks[nbid].Notes[nid].ModifiedAt;
                            }

                            if (NoteMode === "view" && nbid === ViewNotebook && nid === ViewNote) {
                                mustRender = true;
                            }
                        }
                    }
                } else {
                    Notebooks = data.notebooks;

                    for (var nbid in data.notebooks) {
                        for (var nid in data.notebooks[nbid].Notes) {
                            if (Notebooks[nbid].Notes[nid].ModifiedAt > NotebooksModified) {
                                NotebooksModified = Notebooks[nbid].Notes[nid].ModifiedAt;
                            }

                            if (NoteMode === "view" && nbid === ViewNotebook && nid === ViewNote) {
                                mustRender = true;
                            }
                        }
                    }
                }

                var mustNavigate = false;
                if (NewNoteMode !== "") {
                    mustNavigate = true;
                    if (NewNoteMode === "view" && NoteMode === "edit") {
                        mustRender = true;
                    }

                    NoteMode = NewNoteMode;
                    ViewNotebook = NewViewNotebook;
                    ViewNote = NewViewNote;

                    NewNoteMode = "";
                    NewViewNotebook = "";
                    NewViewNote = "";
                }

                if (ViewNotebook === "" && ViewNote === "" && Object.keys(data.notebooks).length > 0) {
                    mustRender = true;
                }

                if (mustRender) {
                    render();
                }
                
                if (nowconnected) {
                    connected = true;
                }

                if (mustNavigate) {
                    if (ViewNote === "") {
                        if (history.length > 2) {
                            history.go(-(history.length - 2));
                        } else {
                            window.location.hash = '#';
                        }
                    } else {
                        window.location.hash = '#' + NoteMode + '/' + ViewNotebook + (ViewNote !== '' ? ('/' + ViewNote) : '');
                    }

                    return;
                }
            }
        } catch (e) {
            console.log(e);
        }

        NoteCheckingItem = false;
    };
    socket.onclose = function (e) {
        connected = false;
        if (ReconnectDelay < 0 || reconnectTimeout != null) {
            return;
        }

        var waitTime = ReconnectDelay;
        console.log("Reconnecting in " + ReconnectDelay + " seconds...");
        reconnectTimeout = setTimeout(ConnectStick, waitTime * 1000)

        ReconnectDelay += (ReconnectDelay * 2) + 1;
        if (ReconnectDelay > 10) {
            ReconnectDelay = 10;
        }
    };
}

function viewNote(notebook, note) {
    NoteMode = "view";
    ViewNotebook = notebook;
    ViewNote = note;

    NoteHideDismissed = !noteHasOption('dismissed');

    render();
}

function editNote(notebook, note) {
    NoteMode = "edit";
    ViewNotebook = notebook;
    ViewNote = note;

    render();
}

function deleteNote(notebook, note) {
    //$('#status').css('display', 'block');

    NewNoteMode = "view";
    NewViewNotebook = "";
    NewViewNote = "";
    payload = {command: "delete", notebook: notebook, note: note};
    socket.send(JSON.stringify(payload));
}

function saveNote(modified) {
    //$('#status').css('display', 'block');

    if (ViewNotebook === "") {
        $('#status').css('display', 'none');
        alert('Failed to save note');
        return;
    }

    var title = $('#editNoteTitle').val();
    if (title === "" && ViewNote === "") {
        alert('Please enter a title');
        return;
    }
    var body = $('#editNoteBody').val();

    NewNoteMode = "view";
    NewViewNotebook = ViewNotebook;
    NewViewNote = ViewNote;

    payload = {command: "edit", notebook: ViewNotebook, note: ViewNote, modified: modified, title: title, body: body};
    socket.send(JSON.stringify(payload));
}

function checkItem(item, m) {
    //$('#status').css('display', 'block');

    if (NoteCheckingItem) {
        setTimeout(function () {
            checkItem(item, m);
        }, 250);
        return;
    }
    NoteCheckingItem = true;

    if (!webSocketReady()) {
        ResumeStick();
        waitForSocketConnection(socket, function () {
            finishCheckItem(item, m);
        });
        return;
    }

    finishCheckItem(item, m);
}

function finishCheckItem(item, m) {
    var notebook = Notebooks[ViewNotebook];
    var note = Notebooks[ViewNotebook].Notes[ViewNote];
    if (!notebook || !note) {
        $('#status').css('display', 'none');
        NoteCheckingItem = false;
        alert('Failed to ' + (checked ? 'check' : 'uncheck') + ' ' + $('#chk_' + item).closest('li').text());
        return;
    }

    payload = {command: "check", notebook: ViewNotebook, note: ViewNote, modified: m, item: item};
    socket.send(JSON.stringify(payload));
}

function toggleHideDismissed() {
    NoteHideDismissed = !NoteHideDismissed;
    render();
}

function noteBody() {
    if (NoteMode == "edit" || NoteMode == "options") {
        if (NoteBodyDraft !== undefined && NoteBodyDraft != null && NoteBodyDraft !== "") {
            return NoteBodyDraft;
        }

        var editBody = $('#editNoteBody').val();
        if (editBody !== undefined && editBody != null) {
            return editBody;
        }
    }

    note = (Object.keys(Notebooks).length > 0 && ViewNote !== "") ? Notebooks[ViewNotebook].Notes[ViewNote] : {};
    if (typeof note !== undefined && note != null && note.Body !== undefined && note.Body != null) {
        return note.Body;
    }

    return "";
}

function noteOptions() {
    body = noteBody();
    if (hasOptionsDirective(body)) {
        var endOfDirective = body.indexOf(')');
        if (endOfDirective > 0) {
            return body.substr(9, endOfDirective - 9).split(',');
        }
    }

    return [];
}

function hasOptionsDirective(body) {
    if (body.substr(0, 9) === '[//]: # (') {
        var endOfDirective = body.indexOf(')');
        if (endOfDirective > 0) {
            var nextNewline = body.indexOf('\n');
            return nextNewline == -1 || nextNewline > endOfDirective;
        }
    }

    return false
}

function noteHasOption(opt) {
    return $.inArray(opt, noteOptions()) != -1;
}

function editNoteOptions() {
    // Store body before changing modes
    NoteBodyDraft = noteBody();

    NoteMode = "options";
    render();
}

function saveNoteOptions() {
    opts = noteOptions();
    for (var i in allNoteOptions) {
        var checked = $('#noteopt' + allNoteOptions[i]).is(':checked');
        var includes = opts.includes(allNoteOptions[i]);
        if (checked && !includes) {
            opts.push(allNoteOptions[i]);
        } else if (!checked && includes) {
            opts = opts.filter(function (opt) {
                return opt !== allNoteOptions[i];
            });
        }
    }
    opts = opts.filter(function (opt) {
        return opt !== "";
    });

    if (!hasOptionsDirective(NoteBodyDraft)) {
        if (opts.length == 0) {
            NoteMode = "edit";
            render();

            return;
        }

        NoteBodyDraft = blankOptionsDirective + NoteBodyDraft;
    }
    var endOfDirective = NoteBodyDraft.indexOf(')');
    if (endOfDirective <= 0) {
        alert('Failed to parse note draft directive');
        return;
    }
    NoteBodyDraft = NoteBodyDraft.substr(0, 9) + opts.join(',') + NoteBodyDraft.substr(endOfDirective);


    NoteMode = "edit";
    render();
}

function render() {
    NoteTitle = "";
    NoteBody = "";
    NoteCheckbox = 0;

    var note;
    if (ViewNotebook === "" || (NoteMode === "view" && ViewNote === "")) {
        NoteTitle = "";

        var sortedNotebooks = [];
        for (var nbid in Notebooks) {
            sortedNotebooks.push([nbid, Notebooks[nbid].Label]);
        }
        sortedNotebooks.sort(sortKeyValuePair);

        NoteBody += '<div class="notebooks">';

        if (Object.keys(Pinned).length > 0) {
            NoteBody += '<div><h2 style="margin-bottom: 4px;display: inline-block;"><small>&#x1F4CC;</small> <span style="text-decoration: underline;">Pinned</span></h2>';

            NoteBody += '<ul style="margin-left: 7px;">';
            for (var nbid in Pinned) {
                for (var i in Pinned[nbid]) {
                    var nid = Pinned[nbid][i];

                    NoteBody += '<li onclick="viewNote(\'' + nbid + '\', \'' + nid + '\')"><h2><a href="#view/' + nbid + '/' + nid + '" class="internal">' + Notebooks[nbid].Notes[nid].Label + '</a></h2></li>';
                }
            }
            NoteBody += '</ul></div>';
        }

        for (let [nbid, nblabel] of sortedNotebooks) {
            NoteBody += '<div><h2 style="margin-bottom: 4px;text-decoration: underline;display: inline-block;"><a onclick="editNote(\'' + Notebooks[nbid].ID + '\', \'\')" class="internal">' + nblabel + '</a></h2>';

            var sortedNotes = [];
            for (var nid in Notebooks[nbid].Notes) {
                sortedNotes.push([nid, Notebooks[nbid].Notes[nid].Label]);
            }
            sortedNotes.sort(sortKeyValuePair);

            NoteBody += '<ul style="margin-left: 7px;">';
            for (let [nid, nlabel] of sortedNotes) {
                NoteBody += '<li onclick="viewNote(\'' + nbid + '\', \'' + nid + '\')"><h2><a href="#view/' + nbid + '/' + nid + '" class="internal">' + nlabel + '</a></h2></li>';
            }
            NoteBody += '<li onclick="editNote(\'' + Notebooks[nbid].ID + '\', \'\')" style="list-style: none;"><h2><a onclick="editNote(\'' + nbid + '\', \'\')" class="internal">+ New</a></h2></li>';
            NoteBody += '</ul></div>';
        }

        NoteBody += '</div>';
    } else {
        note = (Object.keys(Notebooks).length > 0 && ViewNote !== "") ? Notebooks[ViewNotebook].Notes[ViewNote] : {};

        var BackLabel = 'Notebooks';
        var BackAction = 'window.history.back()';
        if (NoteMode !== "view") {
            BackLabel = 'Cancel';
            action = "viewNote";
            if (NoteMode === "options") {
                action = "editNote"
            }
            if (ViewNote !== "") {
                BackAction = action + "('" + ViewNotebook + "', '" + ViewNote + "')";
            } else {
                BackAction = action + "('', '')";
            }
        }
        NoteBody += '<div style="margin-bottom: 7px;"><button onclick="' + BackAction + '">' + BackLabel + '</button> <div style="float: right;">';

        if (NoteMode === "view") {
            if (Object.keys(note).length > 0) {
                toggleHideDismissedLabel = 'Show';
                if (!NoteHideDismissed) {
                    toggleHideDismissedLabel = 'Hide';
                }
                NoteBody += '<button onclick="toggleHideDismissed()">' + toggleHideDismissedLabel + ' &#x2611;</button> &nbsp; ';
            }
            NoteBody += '<button onclick="editNote(\'' + ViewNotebook + '\', \'' + ViewNote + '\')">Edit</button>';
        } else {
            if (NoteMode === "edit") {
                if (note) {
                    NoteBody += '<button onclick="return confirm(\'Are you sure you want to delete ' + note.Label + '?\')?deleteNote(\'' + ViewNotebook + '\', \'' + ViewNote + '\'):false;">Delete note</button> &nbsp;';
                }
                NoteBody += '<button onclick="editNoteOptions();">Options</button> &nbsp;';
            }
            var modified = 0;
            if (Object.keys(Notebooks).length > 0 && ViewNote !== "") {
                modified = Notebooks[ViewNotebook].Notes[ViewNote].ModifiedAt;
            }
            if (NoteMode === "edit") {
                NoteBody += '<button onclick="saveNote(' + modified + ')">Save</button>';
            } else {
                NoteBody += '<button onclick="saveNoteOptions()">Save</button>';
            }
        }
        NoteBody += '</div></div>';

        NoteBody += '<div style="margin-top: 7px;vertical-align: middle;">';
        if (NoteMode === "view") {
            if (note) {
                NoteBody += '<h1 style="display: inline-block;">' + note.Label + '</h1>';
            }
        } else if (NoteMode === "edit") {
            NoteBody += '<input type="text" id="editNoteTitle" placeholder="Title" style="border: 1px dashed #333333;font-size: 1.25em;padding: 7px;width: 100%;box-sizing: border-box;">';
        } else { // Options
            NoteBody += '<h1 style="display: inline-block;">Options</h1>';
        }

        if (NoteMode === "view" && note) {
            modified = new Date(note.ModifiedAt * 1000);
            NoteBody += '<div style="float:right;margin-left: 7px;">&nbsp; ' + modified.getFullYear() + '/' + (modified.getMonth() + 1) + '/' + modified.getDate() + ' ' + ('0' + modified.getHours()).slice(-2) + ':' + ('0' + modified.getMinutes()).slice(-2) + ' &middot; ' + note.ModifiedBy + '</div>';
        }
        NoteBody += '</div>';

        if (Object.keys(note).length > 0 && NoteMode === "view") {
            var noteClass = 'noteNormal';
            if (noteHasOption('list')) {
                noteClass = 'noteList';
            }

            NoteBody += '<div id="noteContent" class="' + noteClass + '">' + marked(note.Body, {
                renderer: markdownRenderer,
                gfm: true,
                breaks: true,
                headerIds: true,
                mangle: false,
                pedantic: false,
                sanitize: false,
                smartLists: true,
                tables: true,
                xhtml: false
            }) + '</div>';
        } else if (NoteMode === "edit") {
            NoteBody += '<textarea id="editNoteBody" placeholder="Text"></textarea>';
        } else { // Options
            opts = noteOptions();

            NoteBody += '<div id="noteContent" class="noteList"><ul>';
            for (let i in allNoteOptions) {
                NoteBody += '<li><label><input type="checkbox" id="noteopt' + allNoteOptions[i] + '" name="noteopt' + allNoteOptions[i] + '"';
                if (opts.includes(allNoteOptions[i])) {
                    NoteBody += ' checked';
                }
                NoteBody += '> ' + allNoteOptionsLabels[i] + '</label></li>';
            }
            NoteBody += '</ul></div>';
        }
    }

    $("#note-body").html(NoteBody);
    if (NoteMode === "view") {
        if (ViewNote !== "" && NoteHideDismissed) {
            var foundChecked = false;
            $('#noteContent').append('<div id="checkedDivider" style="display: none;"><ul id="hiddenItems"></ul></div>').find('input[type=checkbox]:checked').each(function () {
                var closestLi = $(this).closest('li');
                if (!closestLi) {
                    return;
                }

                if (closestLi.find('ul').find('input[type=checkbox]:not(:checked)').length > 0) {
                    return;
                }

                foundChecked = true;
                closestLi.detach().appendTo("#hiddenItems");
            });
            if (foundChecked) {
                $('#checkedDivider').show();
            }
        }
    } else if (NoteMode === "edit") {
        if (NoteBodyDraft != "") {
            $('#editNoteBody').val(NoteBodyDraft);
            NoteBodyDraft = "";
        } else {
            if (note) {
                $('#editNoteBody').val(note.Body);
            }
        }

        bindEditor();
    }

    $('li').each(function (i, li) {
        labelCheckboxes(li);
    });
}

function labelCheckboxes(element) {
    var el = $(element);

    var chli = el.children("ul").children("li");
    if (chli && chli.length > 0) {
        chli.each(function (i, li) {
            labelCheckboxes(li);
        });
    }

    var html = el.html();

    html = '<label>' + html;
    var nextUL = html.indexOf('<ul');
    if (nextUL > -1) {
        html = html.substr(0, nextUL) + '</label>' + html.substr(nextUL);
    } else {
        html += '</label>';
    }

    el.html(html);
}

function bindEditor() {
    $('#editNoteBody').keyup(function (e) {
        if (e.keyCode !== 13 || e.shiftKey) {
            return false;
        }

        var editNoteBody = $("#editNoteBody")[0];
        var selectionStart = editNoteBody.selectionStart;
        var bodyValue = editNoteBody.value;

        var searchLastLineBreak;
        if (selectionStart > 2) {
            searchLastLineBreak = selectionStart - 2;
        } else {
            searchLastLineBreak = +Infinity;
        }

        var lastLineBreak = bodyValue.lastIndexOf("\n", searchLastLineBreak);
        if (lastLineBreak < 0) {
            lastLineBreak = 0
        }

        var m = bodyValue.substr(lastLineBreak, selectionStart - 1 - lastLineBreak).match(regexpListItem);
        if (!m) {
            return false;
        }

        var prefix = m[0];

        var pchk = prefix.substr(prefix.length - 3);
        if (pchk === "[x]" || pchk === "[X]") {
            prefix = prefix.substr(0, prefix.length - 3) + "[ ]";
        }

        prefix = prefix.replace(/^\n+| +$/g, '') + " ";

        editNoteBody.value = bodyValue.substr(0, selectionStart) + prefix + bodyValue.substr(selectionStart);
        editNoteBody.selectionStart = selectionStart + prefix.length;
        editNoteBody.selectionEnd = selectionStart + prefix.length;

        return false;
    });
}

function setAuthKey(newAuthKey) {
    if (newAuthKey == "") {
        newAuthKey = "public";
    }

    AuthKey = newAuthKey;

    db.transaction('rw', db.credentials, async () => {
        db.credentials.clear();
        db.credentials.add({authkey: AuthKey});
    });
}

function processHash() {
    if (window.location.hash.substring(0, 7) !== '#login/' && !haveNoteData()) {
        waitForSocketConnection(socket, function () {
            processHash();
        });
        return;
    }

    if (window.location.hash === '' || window.location.hash === '#') {
        viewNote('', '');
    } else {
        if (window.location.hash.substring(0, 7) === '#login/') {
            var newAuthKey = window.location.hash.substring(7);
            if (newAuthKey !== AuthKey) {
                setAuthKey(newAuthKey);

                // Reset state
                Pinned = {};
                Notebooks = {};
                NotebooksModified = 0;
                NoteMode = "view";
                ViewNotebook = "";
                ViewNote = "";
                render();
                PauseStick();
                ResumeStick();
            }
        } else if (window.location.hash.substring(0, 6) === '#view/' || window.location.hash.substring(0, 6) === '#edit/') {
            var v = window.location.hash.substring(6).split('/');
            var notebookID = v[0];
            var noteID = '';
            if (v.length > 1) {
                noteID = v[1];
            }

            if (window.location.hash.substring(0, 6) === '#view/') {
                if (noteID === '') {
                    notebookID = '';
                }

                viewNote(notebookID, noteID);
            } else { // edit
                editNote(notebookID, noteID);
            }
        }
    }
}

function webSocketReady() {
    return (socket !== null && socket.readyState === 1);
}

function haveNoteData() {
    return webSocketReady && connected == true;
}

function waitForSocketConnection(socket, callback) {
    setTimeout(function () {
        if (webSocketReady()) {
            if (callback != null) {
                callback();
            }
        } else {
            waitForSocketConnection(socket, callback);
        }
    }, 250);
}

function sortKeyValuePair(a, b) {
    return a[1].localeCompare(b[1]);
}

function uppercaseFirst(string) {
    return string.charAt(0).toUpperCase() + string.slice(1);
}
